/*  RDS RADIO CLOCK DUDELFUNK using 
 *  RDM5807M Breakout, Arduino UNO R3, Levelshifter
 *  Datasheet RDM5807M /RDM5803
 *  RDS decoding 2012 at: hjberndt.de/soft/ardsi4735.html
 *  Multi_OLED https://github.com/bitbank2/Multi_OLED
 *  Multi_BitBang https://github.com/bitbank2/Multi_BitBang 
 *  Arduino Clock https://github.com/PaulStoffregen/Time
 *  Arduino TimeZone 
 *  
 *  First published 2021/04/01 at hjberndt.de/soft/rdsradioclock.html
 *  
 */

// OLED MULTI BITBANG
#include <Multi_BitBang.h>
#include <Multi_OLED.h>

#define NUM_DISPLAYS 2
#define NUM_BUSES 1
// I2C bus info
uint8_t scl_list[NUM_BUSES] = {5};
uint8_t sda_list[NUM_BUSES] = {6};
int32_t speed_list[NUM_BUSES] = {1000000L};
// OLED display info
uint8_t bus_list[NUM_DISPLAYS] = {0,0}; // can be multiple displays per bus
uint8_t addr_list[NUM_DISPLAYS] = {0x3c, 0x3d};
uint8_t type_list[NUM_DISPLAYS] = {OLED_132x64, OLED_132x64};
uint8_t flip_list[NUM_DISPLAYS] = {0,0};
uint8_t invert_list[NUM_DISPLAYS] = {0,0};

// TIMELIB TIMEZONE ARDUINO CLOCK
#include <TimeLib.h>
#include <Timezone.h> 
TimeChangeRule CEST = {"", Last, Sun, Mar, 2, 120};     
TimeChangeRule CET = {"", Last, Sun, Oct, 3, 60}; 
Timezone CE(CEST, CET);
TimeChangeRule *tcr; 
char *days[]={"Sonntag",
              "Montag",
              "Dienstag",
              "Mittwoch",
              "Donnerstag",
              "Freitag",
              "Sonnabend"};
              
//RDA RADIO CHIP
#include <Wire.h> //I2C
#define ADDR 0x11
#define WDR5 888
#define DLF 1028
#define ANTENNE 1042
#define STRONGSTATION ANTENNE

// GLOBAL RADIO VARS
word data[7]; //Register 0A-10 RadioRead
boolean Stereo, RDS_Ready, Tune_Complete, Seek_Fail, RDS_Sync, FM_TRUE, FM_READY;
byte    Channel, RSSI, BLERA,BLERB,BLERC,BLERD;

// GLOBAL RDS VARS
char Station[9]="Radiuino";
char RadioText[65]="RadioTXT                                                        ";
char Time[12]="UTC  /Time ";
char Date[11]="Date      ";
word PID,pids; //Counter
word ABFlag=0,r;
bool AFflag=false;
byte R[16];// Response is 12 bytes
word A,B,C,D; //RDS Blocks
word PTY,TP,GT,AB,TA,MS,C12,T16;

/* ---------------------------------------------------------------*/
int dudelfunk = 0;

word getRegister(byte reg)
{Wire.beginTransmission(ADDR); // Device 0x11 for random access
 Wire.write(reg);             
 Wire.endTransmission();      
 Wire.requestFrom(ADDR,2);     
 word x = word(Wire.read(),Wire.read());         
 Wire.endTransmission();   
 return x;
}

void setRegister(byte reg,word val)
{Wire.beginTransmission(ADDR);
 Wire.write(reg); Wire.write(val >> 8); Wire.write(val & 0xFF);
 Wire.endTransmission();
 delay(50);
}

void setFrequency(int freq) //88.8 as 888
{word ch = freq-870;  // Reg 0x03 10 bit 15:6
 ch = ch << 6;        // SHIFT
 bitSet(ch,4);        // TUNE ON
 setRegister(0x03,ch); 
}

word getFrequency(void)
{radioRead();
 return Channel+870;
}

int radioRead() // READ REGISTER 0A-10
{Wire.requestFrom(ADDR-1, 14); //Sequence Addr
 for (int i=0; i<7; i++) { data[i] = 256*Wire.read ()+Wire.read(); }
 Wire.endTransmission();
 getStatus();
}

void getStatus()
{RDS_Ready      = data[0] >> 15 & 1;
 Tune_Complete  = data[0] >> 14 & 1;
 Seek_Fail      = data[0] >> 13 & 1;
 RDS_Sync       = data[0] >> 12 & 1;
 Stereo         = data[0] >> 10 & 1;
 Channel        = data[0] & 0b111111111;
 RSSI           = (data[1] >> 10) & 0x3F;//0..63
 FM_TRUE        = data[1] >> 8 & 1;
 FM_READY       = data[1] >> 7 & 1; 
 BLERA          = data[1] >> 2 & 3;
 BLERB          = data[1] >> 0 & 3;
 BLERC          = data[6] >> 14 & 3;
 BLERD          = data[6] >> 12 & 3;
 //PID            = data[2];
}

void radioInit()
{setRegister(0x02,0b1100000000001011); //RESET
 setRegister(0x03,0b0000000000000000);
 setRegister(0x04,0b0000101000000000);
 setRegister(0x05,0b1000100000001111); 
 setRegister(0x06,0b0000000000000000);
 setRegister(0x07,0b0100001000000010);
 setRegister(0x02,0b1100000000001101); //NO RESET
}

void setup()
{Serial.begin(9600);
 pinMode(8, OUTPUT);

 delay(200);
 Wire.begin(); 
 setRegister(0x02,0xC00d);  
 setFrequency(STRONGSTATION);
 delay(100);
 Serial.print(getFrequency()/10.0);
 Serial.println(" MHz");
 // OLED 0 in use only
 Multi_I2CInit(sda_list, scl_list, speed_list, NUM_BUSES);
 Multi_OLEDInit(bus_list, addr_list, type_list, flip_list, invert_list, NUM_DISPLAYS);
 Multi_OLEDFill(0, 0);
 Multi_OLEDSetContrast(0, 255);
//setTime(13,29,57,2,1,2021);//CHECK BLOCKER 
}

#define MUTE   setRegister(0x02,0b1000000000001101)
#define UNMUTE setRegister(0x02,0b1110000000001101)//Mono bit 13, Mute bit 14

void loop()
{char s[80],tmp[20];static char oldTime[12]; static int last,led=0,tm=0; 
 if(Serial.available()>0)
  {switch(Serial.read())
   {case 'D':setFrequency(DLF);break;
    case 'W':setFrequency(WDR5);break;
    case 'A':setFrequency(ANTENNE);break;
    case 'M':MUTE;
             break;//mute
    case 'm':UNMUTE;
             break;//unmute
              
    case '4'://SEEK DOWN        xxxxxx0xxxxxxxxx
             setRegister(0x02,0b1100000100001101);
             break;//DOWN BIT 9
    case '0':Serial.println(RSSI);
             break;
    case '6'://SEEK UP          xxxxxx1xxxxxxxxx 
             setRegister(0x02,0b1100001100001101);   
             break;  
    case 'f':Serial.println(getFrequency()/10.0);
             break;
    case 'x':sprintf(s,"f=%d RSSI=%d %s Tune_Complete=%d FM_READY=%d",Channel+870,RSSI,Stereo?"Stereo":"Mono",Tune_Complete,FM_READY);
             Serial.println(s);
             sprintf(s,"PI=%04X RDS=%d BLERA=%d BLERB=%d BLERC=%d BLERD=%d",data[2],RDS_Ready,BLERA,BLERB,BLERC,BLERD);
             Serial.println(s);
             break;
    default: Serial.print(Time);Serial.print('\t');
             Serial.print(Date);Serial.print('\t');
             Serial.print(PID,HEX);Serial.print('/');Serial.print(RSSI);Serial.print('/');
             Serial.print(BLERA+BLERB+BLERC+BLERD);Serial.print(' ');Serial.println(Station);
             Serial.println(RadioText);
  }
 }   
 if(analogRead(A1)>500){delay(100);setRegister(0x02,0b1100001100001101); } 
 if(analogRead(A2)>500){delay(100);led=!led;tm=0; } 
 
 
 strcpy(oldTime,Time);
 RDS();getStatus();
 if(strcmp(Time,oldTime))
  if(Station[0]!=Station[7])
   if(BLERA+BLERB+BLERC+BLERD==0)
   {Serial.print("OK ");Serial.print('\t');
    Serial.print(Time);Serial.print('\t');
    Serial.print(Date);Serial.print('\t');
    Serial.print(PID,HEX);Serial.print('/');Serial.print(RSSI);Serial.print('/');
    Serial.print(BLERA+BLERB+BLERC+BLERD);Serial.print(' ');Serial.println(Station);
    Serial.println(RadioText);
    // OLED
    Multi_OLEDFill(0, 0);
    //Multi_OLEDWriteString(0, 0, 0, (char *)"RDS RADIO CLOCK", FONT_NORMAL, 0);
    //CLOCK SYNC
    int hh=atoi(strncpy(s,&Time[0],2));//UTC
    int mm=atoi(strncpy(s,&Time[3],2));
    int dd=atoi(strncpy(s,&Date[0],2));
    int mt=atoi(strncpy(s,&Date[3],2));
    int yy=atoi(strcpy(s,&Date[6]));
    setTime(hh,mm,1,dd,mt,yy); //UTC
    time_t cet=CE.toLocal(now(),&tcr);
    setTime(cet);//TimeZone: UTC --> LOCAL
  
   }
 // OLED LIVE RDS INFO LINES
 int f=getFrequency()/10;int g=getFrequency()%10;
 sprintf(s, "%-8s %3d.%1d %03d %02d", Station,f,g,pids,RSSI); 
 //Serial.println(s);
 Multi_OLEDWriteString(0, 0, 0, s, FONT_SMALL, 0);
 sprintf(s, "%s %s%c%c%c%c%c", strncpy(tmp,Time,5),Date, Stereo?'S':'M',TP?'T':'t',TA?'A':'a',RDS_Ready?'R':'r',RDS_Sync?'S':'s');
 //Serial.println(s);
 Multi_OLEDWriteString(0, 0, 1, s, FONT_SMALL, 0);
 Multi_OLEDWriteString(0, 0, 2, (char *)"---------------------", FONT_SMALL, 0);
 // OLED CLOCK
 
 int ss=second();
 int mm=minute();

 if(ss!=last)  //every second
 {Serial.println(s);
  sprintf(s, "%-10s ",days[weekday()-1]);
  Multi_OLEDWriteString(0, 0, 7, s, FONT_SMALL, 0);
  sprintf(s, "%02d.%02d.%04d",day(),month(),year());
  Multi_OLEDWriteString(0, 66, 7, s, FONT_SMALL, 0);
  sprintf(s, "%2d:%02d:%02d", hour(),minute(),ss);
  Multi_OLEDWriteString(0, 0, 3, s, FONT_LARGE, 0);
  last=ss; 

  //Dudelfunk Blocker: News, Weather, Traffic only
#define TAGS  ((hour()>6 && hour()<18))
#define RUHETAG (feiertag() || day()==1 || day()==7)  
#define VOLL ((mm == 59) && ss > 57)
#define HALB ((mm == 29) && ss > 54) 
#define DURCHSAGE (TA)

if DURCHSAGE {delay(1000); RDS();}//anti flatter?

  if VOLL led = 1;
  if (!RUHETAG && TAGS && HALB)led = 1;
  if DURCHSAGE led = 2;
  if(led == 2 && !DURCHSAGE){led=0;tm=0;}
  
  if(led==1)tm++; if(tm>300){led=0;tm=0;} //5 minutes to mute
  
  if(led>0)UNMUTE; 
        else MUTE;
  digitalWrite(8,led>0);
  
  sprintf(s,"%d%c%c%c%c %03d",led,
                 RUHETAG?'R':'r',
                 TAGS?'T':'t',
                 VOLL?'V':'v',
                 HALB?'H':'h',tm);
  Serial.println(s);
 }
 delay(100);
}

/* RDS DECODER FROM Arduino goes Radio hjberndt.de 2012
 * hjberndt.de/soft/ardsi4735.html 
 * COPY data to old R[] ist the only alg. change.  
 * Global vars are used
 */

int RDS() //Page 77
{int i,loopcount=1;
 while(loopcount--)
 {// added lines
  radioRead();
  R[4]=data[2]/256;
  R[5]=data[2]&255;
  R[6]=data[3]/256;
  R[7]=data[3]&255;
  R[8]=data[4]/256;
  R[9]=data[4]&255;
  R[10]=data[5]/256;
  R[11]=data[5]&255;
 
  // end added lines
  A = data[2];//MAKEINT(R[ 4],R[ 5]); //build BLOCKS
  B = data[3];//MAKEINT(R[ 6],R[ 7]);
  C = data[4];//MAKEINT(R[ 8],R[ 9]);
  D = data[5];//MAKEINT(R[10],R[11]);
  //Program Identification PID
  //PID change?
  if(PID!=A)
  {for(i=0;i<64;i++)RadioText[i]=32;
   for(i=0;i< 8;i++)Station[i]=32;
   pids=0; 
  }
  if(pids<999)pids++; // PID counter
  PID = A; 
  PTY = (B>>5)   &0x1F; //0 - 31 -> 10 Pop
  TP  = (R[6]>>2)&0x01; // Traffic Program
  GT  = (R[6]>>4)&0x0F; // Group Type 0 - 15
  AB  = (R[6]>>3)&0x01; // Group A/B -> 0/1
 // TA = 0;
 // MS = 0;
 // C12= 0;
 // T16= 0;
//  if ( (R[6] & 248) == 64 ){strcpy(text,"4");return GT;}
  Station[8]='\0';
  RadioText[64]='\0';
  
  switch(AB)
  {case 0:// Type A 
    
    switch(GT)// GroupType
        {case 0: //Station callsign
         TA = bitRead(B,4);// Block B Bit 4 OK
         MS = (R[7]>>3)&0x01;//MS = bitRead(B,3);
         C12= R[7]&3;
        
         switch(C12)
         {case 0:Station[0]=R[10]; Station[1]=R[11]; break;
          case 1:Station[2]=R[10]; Station[3]=R[11]; break;
          case 2:Station[4]=R[10]; Station[5]=R[11]; break;
          case 3:Station[6]=R[10]; Station[7]=R[11]; break;
         }
         //AF
         //if ((R[8]>224) && (R[8]<250))AFn = R[8] - 224;
         r=(AFflag?R[8]:R[9]);
     r=PID;    
         AFflag!=AFflag;
         break;
     case 2: //RADIOTEXT
         i= (R[7]&0xF0); //ClearLine? Bit 4
         if (ABFlag!=i)for(ABFlag=i,i=0;i<64;i++)RadioText[i]=32;
         T16 = R[7]&0x0F;//RadioText  
         if(T16==0)loopcount=32;
         if (loopcount) delay(2);
         RadioText[T16*4+00]=R[ 8];
         RadioText[T16*4+01]=R[ 9];
         RadioText[T16*4+02]=R[10];
         RadioText[T16*4+03]=R[11];
       for(i=0;i<64;i++)if(RadioText[i]<' ')RadioText[i]='*';
         break;
     case 4: //GroupType 4A: Time every minute
         {int UTCoffset = (R[11] & 31) / 2;
          if ( (R[11] & 32) == 32 )UTCoffset = UTCoffset * -1;
          int mm = R[11] / 64 + 4 * (R[10] & 15);
          int hh = R[10] / 16 + 16 * (R[9] & 1);
          int hhLocal = hh + UTCoffset;
          if ( hhLocal > 23 )hhLocal -= 24;
          sprintf(Time, "%02d:%02d/%02d:%02d",hh,mm,hhLocal,mm);
         }
         //Date UTC
         {unsigned long MJD = (C>>1) | ((R[7]&3)<<15);
          word Y=(MJD-15078.2)/365.25;
          word M=(MJD-14956.1-(word)(Y*365.25))/30.6001;    
          byte K; word YY,MM,DD;
          if(M==14||M==15)K=1;
          else  K=0;
          YY=(Y+K)%100;
          MM=M-1-K*12;
          DD=MJD-14956-(word)(Y*365.25)-(word)(M*30.6001);
          sprintf(Date, "%02d.%02d.%04d",DD,MM,YY+2000);
         }
         break;    
     default: break;
    }//switch GT
    break;
   case 1: //Type B not implemented due to no reception
        
    switch(GT)// GroupType
    {case 2: //T16 = R[7]&0x0F;//RadioText  
         //RadioText[T16*2+00]=R[10];
         //RadioText[T16*2+01]=R[11];
         break;
     default:break;    
    }
        break;   
   default:break;
  }// switch AB
 }//loopcount
  //for(i=0;i<63;i++)if(RadioText[i]<32)RadioText[i]='_';
  return r;
}


// unverändert aus Arduino Forum "Feiertagsberechnung"
// https://forum.arduino.cc/index.php?topic=320217.0

boolean feiertag()
{
  String testDate;
  int tag;
  int monat=3;
  int ostersonntag;
  
  testDate = String(day()) +"."+ String(month());

        // Zuerst die festen Feiertage
        // Hinweis: 24.12 und 31.12 sind eigentlich keine Feiertage, werden aber hier als solche behandelt
  if (testDate==("1.1")) {return true;}   // Neujahr
  if (testDate==("1.5")) {return true;}   // Maifeiertag
  if (testDate==("3.10")) {return true;}  // Tag d. dt. Einheit
  if (testDate==("24.12")) {return true;} // Heiligabend (eigentlich kein Feiertag)
  if (testDate==("25.12")) {return true;} // Erster Weihnachtsfeirtag
  if (testDate==("26.12")) {return true;} // Zweiter Weihnachsfeiertag
  if (testDate==("31.12")) {return true;} // Silvester (eigentlich kein Feiertag)

        // Nachfolgend einige Feiertage die nicht bundeseinheitlich sind: (ggf. auskommentieren)
  // if (testDate==("6.1")) {return true;}   // Heilige Drei KÃ¶nige (nur in best. BundeslÃ¤ndern)
  // if (testDate==("15.8")) {return true;}  // Mariae Himmelfahrt (nur im Saarland)
  // if (testDate==("31.10")) {return true;} // Reformationstag (nur in best. BundeslÃ¤ndern)
  // if (testDate==("1.11")) {return true;}  // Allerheiligen (nur in best. BundeslÃ¤ndern)



   if (month()==3) {tag=day();}  // Wenn MÃ¤rz, aktuellen Tag ermitteln
   if (month()==4) {tag=day()+31;} 
   if (month()==5) {tag=day()+31+30;} 
   if (month()==6) {tag=day()+31+30+31;} 
   
   ostersonntag = berechne_Ostern();
   if (ostersonntag==tag)  {return true;}     // Ostersonntag
   if (ostersonntag-2==tag)  {return true;}   // Karfreitag
   if (ostersonntag+1==tag)  {return true;}   // Ostermontag
   if (ostersonntag+39==tag)  {return true;}  // Christi Himmelfahrt
   if (ostersonntag+49==tag)  {return true;}  // Pfingstsonntag
   if (ostersonntag+50==tag)  {return true;}  // Pfingstmontag
   if (ostersonntag+60==tag)  {return true;}  // Fronleichnam (nicht bundeseinheitlich)

   // Buss- und Bettag gibt es nur in eingen BundelÃ¤ndern, also ggf. auskommentieren
   // if (buss_und_bettag()) {return true;}

return false;
}

  
int berechne_Ostern()
{
  // nach der GauÃŸ-Formel
  // RÃ¼ckgabewert: Datum von Ostern ab 1. MÃ¤rz (max=56 fÃ¼r 25. April)

  int a;
  int b;
  int c;
  int d;
  int e;
  int f;
  int k;
  int M;
  int N;
  int p;
  int q;
  int jahr=year();
  

  // Die "magische" Gauss-Formel anwenden:
  a = jahr % 19;
  b = jahr % 4;
  c = jahr % 7;
  k = jahr / 100;
  p = (8*k+13) / 25;
  q = k / 4;
  M =(15 + k - p - q) % 30;
  N = (4 + k-q) % 7;
  d = (19*a + M) % 30;
  e = (2*b + 4*c + 6*d + N) % 7;
  f = 22+d+e;   // Tag auf Maerz bezogen: 32 bedeutet Erster April usw.

  
  if (f==57) {f=50;}   // Wenn der 26. April ermittelt wird (31+26=57), dann muss Ostern am 19.4. sein (19+31=50)
  /// Falls der 25. April ermittelt wird, gilt dies nur wenn d=28 und a>10
  if ((f==56) && (d==28) && (a>10))  {f=49;}
  return f;
  
 }

boolean buss_und_bettag()
{
  // Buss- und Bettag fÃ¤llt auf den Mittwoch zwischen demm 16. und 22. November
  if ( (month()==11) && (day()>15) && (day ()< 23) )
    {
      if (weekday() == 4) { return true; }  // Wenn heute Mittwoch ist
     }
      
  return false;
      
}
